#include <stdio.h>
#include <math.h>

/* maximum number of iterations */
#define	 MAXLOOP		50

/* definition of the macro ROTATE */
#define	 ROTATE(a,i,j,k,l)	g=a[i][j];h=a[k][l];a[i][j]=g-s*(h+g*tau);\
				a[k][l]=h+s*(g-h*tau)

/* 
 * Jacobi Transformation of a Symmetric Matrix
 *
 * Purpose
 *   Computes all eigenvalues and eigenvectors of a real symmetric 
 *   matrix a[n][n].
 *
 * Input
 *   a[n][n] is the n x n symmetric matrix that is allocated by 
 *	the function allocate_matric()
 *   n is the number of rows and columns
 *
 * Output
 *   d[n] contains eigenvalues of a.
 *   v[n][n] is a matrix whose columns contain the normalized
 *   	eigenvectors of a.
 *   nrot returns the number of Jacobi rotations that were required.
 */

void
jacobi (a, n, d, v, nrot)
double	*a[], *d, **v;
int	n, *nrot;

{
    int    j, iq, ip, i;
    double tresh, theta, tau, t, sm, s, h, g, c, *b, *z;

    /* allocate memory for vectors */
    b = (double *) malloc (n * sizeof(double));
    z = (double *) malloc (n * sizeof(double));

    /* initialize the identity matrix */
    for (ip = 0; ip < n; ip++) {
	/* clear the matrix */
	for (iq = 0; iq < n; iq++)
	    v[ip][iq] = 0.0;
	
	/* set the diagonal to 1 */
	v[ip][ip] = 1.0;
    }

    /* initlaize b and d to the diagonal of a */
    for (ip = 0; ip < n; ip++) {
        b[ip] = d[ip] = a[ip][ip];
        z[ip] = 0.0;
    }
    *nrot = 0;

    /* main loop of Jacobi Transformation */
    for (i = 0; i < MAXLOOP; i++) {
	/* calculate the sum of off-diagonal elements */
	for (sm = 0.0, ip = 0; ip < n-1; ip++) {
	    for (iq = ip+1; iq < n; iq++)
		sm += fabs(a[ip][iq]);
	}

	/* check the result of sm */
	if (sm == 0.0) {
	    /* the normal return, which relies on quadratic convergence
	       to machine underflow */
	    free (b);
	    free (z);
	    return;
	}

	/* In the first sweeps, we carry out the pq rotation only
	   if |apq| > e for some threshold value */
	if (i < 3)
	   tresh = 0.2 * sm / (n*n);
	else
	   tresh = 0.0;
	
	/* sweep */
	for (ip = 0; ip < n-1; ip++) {
	    for (iq = ip+1; iq < n; iq++) {
		g = 100.0 * fabs(a[ip][iq]);

		/* after 4 sweeps, skip the rotation 
		   if the off-diagonal is small */
		if ((i >= 4) && (fabs(d[ip])+g == fabs(d[ip]))
		    && (fabs(d[iq])+g == fabs(d[iq])))
		    a[ip][iq] = 0.0;
		else
		if (fabs(a[ip][iq]) > tresh) {
		    h = d[iq] - d[ip];
		    if (fabs(h) + g == fabs(h))
			t = a[ip][iq] / h;
		    else {
			theta = 0.5 * h / a[ip][iq];
			t = 1.0 / (fabs(theta) + sqrt (1.0+theta*theta));
			if (theta < 0.0)
			    t = -t;
		    }
		    c = 1.0 / sqrt(1+t*t);
		    s = t * c;
		    tau = s / (1.0+c);
		    h = t * a[ip][iq];
		    z[ip] -= h;
		    z[iq] += h;
		    d[ip] -= h;
		    d[iq] += h;
		    a[ip][iq] = 0.0;
		    for (j = 0; j < ip; j++) {
			ROTATE(a, j, ip, j, iq);
		    }
		    for (j = ip+1; j < iq; j++) {
			ROTATE(a, ip, j, j, iq);
		    }
		    for (j = iq+1; j < n; j++) {
			ROTATE(a, ip, j, iq, j);
		    }
		    for (j = 0; j < n; j++) {
			 ROTATE(v, j, ip, j, iq);
		    }
		    (*nrot)++;
		}
	    }
	}

	/* update d and reinitialize z */
	for (ip = 0; ip < n; ip++) {
	    b[ip] += z[ip];
	    d[ip] = b[ip];
	    z[ip] = 0.0;
	}
    }
    fprintf(stderr, "Too many iterations in routine Jacobi\n");
    free (b);
    free (z);
}

/*
 * Allocate memory for a matrix
 */

void	**
allocate_matrix (row, column, size)
int    row;			/* number of rows */
int    column;			/* number of columns */
int    size;			/* size of each element */

{
    int    i;
    void   **m;

    /* allocate memory for pointers to rows */
    m = (void **) malloc (row*sizeof(void *));

    /* allocate memory for columns */
    for (i = 0; i < row; i++)
        m[i] = (void *) malloc (column*size);

    /* return the allocated matrix */
    return (m);
}

/*
 * Free allocated matrix
 */

void
free_matrix (row, column, m)
int	row;
int	column;
void	**m;			/* allocated matrix */

{
    int	  i;

    /* free allocated memory */
    for (i = 0; i < row; i++)
	free (m[i]);
    free (m);
}
